@@ -1,15 +1,21 @@
+import os
+
 import requests
 import json
-
+import logging
 from pathlib import Path
-from typing import List, Dict
+from typing import List, Dict, BinaryIO
 from .endpoints import Endpoint
 from .models import Project, Dataset, Image
 from datagym.exceptions.exceptions import (APIException,
                                            InvalidTokenException,
                                            ClientException,
+                                           ClientExceptionNonFatal,
                                            ExceptionMessageBuilder)
 
+from datagym.utils.loadingbar import progressbar
+from datagym.constants import WARNING_KEYS
+
 
 class Client:
     """The Client class provides convenient access to DataGym's API.
@@ -26,6 +32,7 @@ class Client:
     """
 
     MAX_NUM_URLS_PER_UPLOAD = 50
+    logger = logging.getLogger(__name__)
 
     def __init__(self, api_key: str) -> None:
         """ Initializes DataGym Client instance
@@ -60,14 +67,16 @@ class Client:
             method: str,
             endpoint: str,
             headers: dict or None,
-            data: dict or None
+            json: list or dict = None,
+            data: BinaryIO or dict = None
     ) -> requests.Response:
         """ Send a HTTP request to a DataGym endpoint
 
         :param str method: The HTTP method (ex. GET, POST, PUT, etc.)
         :param str endpoint: The DataGym API endpoint
         :param dict headers: The HTTP request header
-        :param dict data: The request body as json
+        :param dict json: The request body as json
+        :param dict data: The request body
 
         :raises requests.exceptions.ConnectionError: Connection errors
 
@@ -75,10 +84,20 @@ class Client:
         :rtype: requests.Response
         """
         try:
-            return requests.request(method=method,
-                                    url=self._endpoint.BASE_PATH + endpoint,
-                                    headers=headers,
-                                    json=data)
+            if json:
+                return requests.request(method=method,
+                                        url=self._endpoint.BASE_PATH + endpoint,
+                                        headers=headers,
+                                        json=json)
+            elif data:
+                return requests.request(method=method,
+                                        url=self._endpoint.BASE_PATH + endpoint,
+                                        headers=headers,
+                                        data=data)
+            else:
+                return requests.request(method=method,
+                                        url=self._endpoint.BASE_PATH + endpoint,
+                                        headers=headers)
         except requests.exceptions.ConnectionError as e:
             raise e
         except requests.exceptions.RequestException as e:
@@ -93,24 +112,24 @@ class Client:
         """
         if response.ok:
             return True
+        elif json.loads(response.content)["key"] in WARNING_KEYS:
+            ClientExceptionNonFatal(self._msg_builder, self.logger, **json.loads(response.content))
+            return False
         elif response.status_code == 500:
             if response.content:
-                raise APIException(self._msg_builder,
-                                   **json.loads(response.content))
+                raise APIException(self._msg_builder, **json.loads(response.content))
             else:
                 response.raise_for_status()
         else:
             if response.content:
-                raise ClientException(self._msg_builder,
-                                      **json.loads(response.content))
+                raise ClientException(self._msg_builder, **json.loads(response.content))
             else:
                 response.raise_for_status()
 
     def _is_token_valid(self) -> bool:
         response = self._request(method="HEAD",
                                  endpoint=self._endpoint.PROJECT,
-                                 headers=self.__auth,
-                                 data=None)
+                                 headers=self.__auth)
 
         if not response.ok:
             raise InvalidTokenException(msg_builder=self._msg_builder,
@@ -131,8 +150,7 @@ class Client:
         """
         response = self._request(method="GET",
                                  endpoint=self._endpoint.PROJECT,
-                                 headers=self.__auth,
-                                 data=None)
+                                 headers=self.__auth)
 
         if self._response_valid(response):
             content = json.loads(response.content)
@@ -178,8 +196,7 @@ class Client:
         """
         response = self._request(method="GET",
                                  endpoint=self._endpoint.DATASET,
-                                 headers=self.__auth,
-                                 data=None)
+                                 headers=self.__auth)
 
         if self._response_valid(response):
             content = json.loads(response.content)
@@ -213,8 +230,7 @@ class Client:
 
         response = self._request(method="GET",
                                  endpoint=endpoint,
-                                 headers=self.__auth,
-                                 data=None)
+                                 headers=self.__auth)
 
         if self._response_valid(response):
             return json.loads(response.content)
@@ -231,8 +247,7 @@ class Client:
 
         response = self._request(method="HEAD",
                                  endpoint=endpoint,
-                                 headers=None,
-                                 data=None)
+                                 headers=None)
 
         if self._response_valid(response):
             return response.url
@@ -251,8 +266,7 @@ class Client:
 
         response = self._request(method="GET",
                                  endpoint=endpoint,
-                                 headers=self.__auth,
-                                 data=None)
+                                 headers=self.__auth)
 
         if self._response_valid(response):
             with open(path_file, 'wb') as handler:
@@ -270,8 +284,7 @@ class Client:
 
         response = self._request(method="GET",
                                  endpoint=endpoint,
-                                 headers=self.__auth,
-                                 data=None)
+                                 headers=self.__auth)
 
         if self._response_valid(response):
             return response.content
@@ -302,10 +315,12 @@ class Client:
         response = self._request(method="POST",
                                  endpoint=self._endpoint.DATASET,
                                  headers=headers,
-                                 data=data)
+                                 json=data)
 
         if self._response_valid(response):
             return Dataset(json.loads(response.content))
+        else:   # In case of non-fatal exception
+            return self.get_dataset_by_name("name")
 
     def add_dataset(self, dataset_id: str, project_id: str) -> bool:
         """ Add a Dataset to a Project
@@ -320,11 +335,12 @@ class Client:
 
         response = self._request(method="POST",
                                  endpoint=endpoint,
-                                 headers=self.__auth,
-                                 data=None)
+                                 headers=self.__auth)
 
         if self._response_valid(response):
             return True
+        else:   # In case of non-fatal exception (already attached)
+            return True
 
     def remove_dataset(self, dataset_id: str, project_id: str) -> bool:
         """ Remove a Dataset from a Project
@@ -339,8 +355,7 @@ class Client:
 
         response = self._request(method="DELETE",
                                  endpoint=endpoint,
-                                 headers=self.__auth,
-                                 data=None)
+                                 headers=self.__auth)
 
         if self._response_valid(response):
             return True
@@ -364,14 +379,18 @@ class Client:
             response = self._request(method="POST",
                                      endpoint=endpoint,
                                      headers=self.__auth,
-                                     data=image_url_list)
+                                     json=image_url_list)
             if self._response_valid(response):
                 return json.loads(response.content)
         # If the url List contains a large number of urls it is split into mini batches for upload
         else:
             response = []
 
-            for i in range(0, len(image_url_list), self.MAX_NUM_URLS_PER_UPLOAD):
+            for i in progressbar(
+                    range(0, len(image_url_list), self.MAX_NUM_URLS_PER_UPLOAD),
+                    "Uploading image urls: ",
+                    40
+            ):
                 if i + self.MAX_NUM_URLS_PER_UPLOAD >= len(image_url_list):
                     slice = image_url_list[i:]
                 else:
@@ -380,7 +399,7 @@ class Client:
                 partial_response = self._request(method="POST",
                                                  endpoint=endpoint,
                                                  headers=self.__auth,
-                                                 data=slice)
+                                                 json=slice)
 
                 if self._response_valid(partial_response):
                     response += json.loads(partial_response.content)
@@ -401,8 +420,7 @@ class Client:
 
         response = self._request(method="DELETE",
                                  endpoint=endpoint,
-                                 headers=self.__auth,
-                                 data=None)
+                                 headers=self.__auth)
 
         if self._response_valid(response):
             return True
@@ -422,7 +440,7 @@ class Client:
             response = self._request(method="POST",
                                      endpoint=endpoint,
                                      headers=self.__auth,
-                                     data=label_data)
+                                     json=label_data)
 
             if self._response_valid(response):
                 return json.loads(response.content)
@@ -430,7 +448,11 @@ class Client:
         else:
             response = []
 
-            for i in range(0, len(label_data), self.MAX_NUM_URLS_PER_UPLOAD):
+            for i in progressbar(
+                    range(0, len(label_data), self.MAX_NUM_URLS_PER_UPLOAD),
+                    "Uploading annotations: ",
+                    40
+            ):
                 if i + self.MAX_NUM_URLS_PER_UPLOAD >= len(label_data):
                     mini_batch = label_data[i:]
                 else:
@@ -439,7 +461,7 @@ class Client:
                 partial_response = self._request(method="POST",
                                                  endpoint=endpoint,
                                                  headers=self.__auth,
-                                                 data=mini_batch)
+                                                 json=mini_batch)
 
                 if self._response_valid(partial_response):
                     response += json.loads(partial_response.content)
@@ -447,3 +469,59 @@ class Client:
                     break
 
             return response
+
+    def upload_image(self, dataset_id: str, image_path: str, image_name: str = None) -> bool:
+        """ Uploads an Image to a Dataset
+
+        :param str dataset_id: The dataset the image should be uploaded to
+        :param str image_path: The path to the image that should be uploaded
+        :param str image_name: Your prefered image name.
+                                If left empty, it will automatically be extracted from the image path
+        :returns: True if Image was successfully uploaded
+        :rtype: bool
+
+        """
+        endpoint = self._endpoint.upload_image(dataset_id)
+
+        if not image_name:
+            image_name = os.path.basename(image_path)
+
+        files = open(image_path, 'rb')
+
+        headers = {
+            'X-filename': image_name,
+            **self.__auth
+        }
+
+        response = self._request(method="POST",
+                                 endpoint=endpoint,
+                                 headers=headers,
+                                 data=files)
+
+        if self._response_valid(response):
+            return Image(json.loads(response.content))
+
+    def upload_label_config(self, config_id: str, label_config: List[Dict]) -> bool:
+        """ Clears the existing config and replaces it by a new one.
+        Careful with this as clearing the config also clears all associated labels
+
+        :param config_id:
+        :param label_config:
+        :return: True if successful
+        """
+        clear_endpoint = self._endpoint.upload_label_config(config_id)
+        upload_endpoint = self._endpoint.clear_label_config(config_id)
+
+        clear_response = self._request(method="DELETE",
+                                       endpoint=clear_endpoint,
+                                       headers=self.__auth)
+
+        if self._response_valid(clear_response):
+
+            upload_response = self._request(method="PUT",
+                                            endpoint=upload_endpoint,
+                                            headers=self.__auth,
+                                            json=label_config)
+
+            if self._response_valid(upload_response):
+                return upload_response.content
